﻿using Microsoft.Diagnostics.Tracing;
using Microsoft.Diagnostics.Tracing.Etlx;
using Microsoft.Diagnostics.Tracing.Parsers;
using Microsoft.Diagnostics.Tracing.Parsers.Clr;
using Microsoft.Diagnostics.Tracing.Session;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Threading;

namespace ConsoleApp1
{
	class ClrEventSourceMonitor
	{
		static TextWriter Out = Console.Out;
		static Dictionary<ulong, string> _types = new Dictionary<ulong, string>();
		static int _processId;

		public static void SetTargetProcessId(int processId)
		{
			_processId = processId;
		}

		public static int Run()
		{
			// Today you have to be Admin to turn on ETW events (anyone can write ETW events).   
			if (!(TraceEventSession.IsElevated() ?? false))
			{
				Out.WriteLine("To turn on ETW events you need to be Administrator, please run from an Admin process.");
				Debugger.Break();
				return -1;
			}

			var sessionName = "SimpleMontitorSession";
			using (var session = new TraceEventSession(sessionName, null))
			{
				var restarted = session.EnableProvider(
					ClrTraceEventParser.ProviderGuid, TraceEventLevel.Verbose,
					(ulong)(ClrTraceEventParser.Keywords.GC
                    | ClrTraceEventParser.Keywords.GCHeapAndTypeNames
					| ClrTraceEventParser.Keywords.Type));

				Console.CancelKeyPress += delegate (object sender, ConsoleCancelEventArgs e) { session.Dispose(); };

				using (TraceLogEventSource traceLogSource = TraceLog.CreateFromTraceEventSession(session))
				{
					// 이벤트 발생: ClrTraceEventParser.Keywords.Type
					//traceLogSource.Clr.TypeBulkType += delegate (GCBulkTypeTraceData data)
					//{
					//	if (data.ProcessID != processId)
					//	{
					//		return;
					//	}

					//	Console.Write(". ");

					//	for (int currentType = 0; currentType < data.Count; currentType++)
					//	{
					//		GCBulkTypeValues value = data.Values(currentType);

					//		// ClrTraceEventParser.Keywords.GCHeapAndTypeNames - value.TypeName
					//		_types[value.TypeID] = value.TypeName;
					//	}
					//};

					// 이벤트 발생: ClrTraceEventParser.Keywords.GC
					traceLogSource.Clr.GCFinalizeObject += delegate (FinalizeObjectTraceData data)
					{
						if (_processId == 0 || data.ProcessID != _processId)
						{
							return;
						}

						/*
						data.TypeName ==> ClrTraceEventParser.Keywords.GCHeapAndTypeNames | ClrTraceEventParser.Keywords.Type
						*/

						//if (_types.ContainsKey(data.TypeID) == false)
						//{
						//	Out.WriteLine($"{data.TimeStamp}, {data.TypeID}, unknown-type, {data.TypeName}");
						//}
						//else
						//{
						//	Out.WriteLine($"{data.TimeStamp}, {data.TypeID}, {_types[data.TypeID]}, {data.TypeName}");
						//}

						Out.WriteLine($"{data.TimeStamp}, {data.TypeID}, {data.TypeName}");
					};

					traceLogSource.Process();
				}

				if (restarted)      // Generally you don't bother with this warning, but for the demo we do. 
					Out.WriteLine("The session {0} was already active, it has been restarted.", sessionName);

				Out.WriteLine("**** Start listening for events");

				session.Source.Process();
				Out.WriteLine();
				Out.WriteLine("Stopping the collection of events.");
			}
			return 0;
		}
	}
}
